@@ -1,3 +1,4 @@
+import functools
 import os
 import plistlib
 import subprocess
@@ -10,7 +11,7 @@ import colander
 
 from .forms import (cancel_job_form,
                     release_job_form,
-                    create_job_subscriptions_form,
+                    create_job_subscription_form,
                     create_printer_subscription_form,
                     cups_add_modify_class_form,
                     cups_add_modify_printer_form,
@@ -26,6 +27,7 @@ from .forms import (cancel_job_form,
                     get_jobs_form,
                     get_printer_attributes_form,
                     get_subscriptions_form,
+                    get_notifications_form,
                     pause_printer_form,
                     resume_printer_form,
                     hold_new_jobs_form,
@@ -33,15 +35,28 @@ from .forms import (cancel_job_form,
                     cancel_subscription_form,
                     )
 
+try:
+    if os.getenv('NO_TORNADO', '').lower() in ('1', 'yes', 'true', 't'):
+        raise ImportError
+    from tornado.gen import coroutine, Return, Task
+    from tornado.ioloop import TimeoutError
+except ImportError:
+    HAS_TORNADO = False
 
-class TimeoutError(Exception):
-    pass
+    def coroutine(f):
+        return f
+
+    class TimeoutError(Exception):
+        pass
+else:
+    HAS_TORNADO = True
 
 
 class IPPToolWrapper(object):
 
-    def __init__(self, config):
+    def __init__(self, config, io_loop=None):
         self.config = config
+        self.io_loop = io_loop
 
     def authenticate_uri(self, uri):
         if 'login' in self.config and 'password' in self.config:
@@ -69,6 +84,42 @@ class IPPToolWrapper(object):
             time.sleep(.1)
 
     def _call_ipptool(self, uri, request):
+        if HAS_TORNADO:
+            return self._async_call_ipptool(uri, request)
+        return self._sync_call_ipptool(uri, request)
+
+    @coroutine
+    def cleanup_fd(self, name):
+        try:
+            os.unlink(name)
+        except OSError:
+            pass
+
+    @coroutine
+    def _async_call_ipptool(self, uri, request):
+        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
+            temp_file.write(request)
+        from tornado.process import Subprocess
+        process = Subprocess([self.config['ipptool_path'],
+                              self.authenticate_uri(uri), '-X',
+                              temp_file.name],
+                             stdin=subprocess.PIPE,
+                             stdout=Subprocess.STREAM,
+                             stderr=Subprocess.STREAM,
+                             io_loop=self.io_loop)
+        future = []
+        self.io_loop.add_timeout(self.io_loop.time() + self.config['timeout'],
+                                 functools.partial(self.timeout_handler,
+                                                   process.proc, future))
+        stdout, stderr = yield [Task(process.stdout.read_until_close),
+                                Task(process.stderr.read_until_close)]
+        if future:
+            raise TimeoutError
+        yield self.cleanup_fd(temp_file.name)
+
+        raise Return(plistlib.readPlistFromString(stdout))
+
+    def _sync_call_ipptool(self, uri, request):
         with tempfile.NamedTemporaryFile(delete=False) as temp_file:
             temp_file.write(request)
         process = subprocess.Popen([self.config['ipptool_path'],
@@ -117,20 +168,20 @@ class IPPToolWrapper(object):
         response = self._call_ipptool(uri, request)
         return response['Tests'][0]
 
-    def create_job_subscriptions(self,
-                                 uri,
-                                 printer_uri=None,
-                                 requesting_user_name=None,
-                                 notify_job_id=None,
-                                 notify_recipient_uri=colander.null,
-                                 notify_pull_method=colander.null,
-                                 notify_events=colander.null,
-                                 notify_attributes=colander.null,
-                                 notify_charset=colander.null,
-                                 notify_natural_language=colander.null,
-                                 notify_time_interval=colander.null):
+    def create_job_subscription(self,
+                                uri,
+                                printer_uri=None,
+                                requesting_user_name=None,
+                                notify_job_id=None,
+                                notify_recipient_uri=colander.null,
+                                notify_pull_method=colander.null,
+                                notify_events=colander.null,
+                                notify_attributes=colander.null,
+                                notify_charset=colander.null,
+                                notify_natural_language=colander.null,
+                                notify_time_interval=colander.null):
         """
-        Create one or more Per-Job Subscription objects.
+        Create a per-job subscription object.
         """
         kw = {'header': {'operation_attributes':
                          {'printer_uri': printer_uri,
@@ -143,7 +194,7 @@ class IPPToolWrapper(object):
               'notify_charset': notify_charset,
               'notify_natural_language': notify_natural_language,
               'notify_time_interval': notify_time_interval}
-        request = create_job_subscriptions_form.render(kw)
+        request = create_job_subscription_form.render(kw)
         response = self._call_ipptool(uri, request)
         return response['Tests'][0]
 
@@ -152,10 +203,14 @@ class IPPToolWrapper(object):
             uri,
             printer_uri=None,
             requesting_user_name=None,
-            notify_recipient_uri=None,
-            notify_events=None,
+            notify_recipient_uri=colander.null,
+            notify_pull_method=colander.null,
+            notify_events=colander.null,
+            notify_attributes=colander.null,
+            notify_charset=colander.null,
+            notify_natural_language=colander.null,
             notify_lease_duration=colander.null,
-            notify_lease_expiration_time=colander.null):
+            notify_time_interval=colander.null):
         """
         Create a new subscription and return its id
         """
@@ -163,9 +218,13 @@ class IPPToolWrapper(object):
                          {'printer_uri': printer_uri,
                           'requesting_user_name': requesting_user_name}},
               'notify_recipient_uri': notify_recipient_uri,
+              'notify_pull_method': notify_pull_method,
               'notify_events': notify_events,
+              'notify_attributes': notify_attributes,
+              'notify_charset': notify_charset,
+              'notify_natural_language': notify_natural_language,
               'notify_lease_duration': notify_lease_duration,
-              'notify_lease_expiration_time': notify_lease_expiration_time}
+              'notify_time_interval': notify_time_interval}
         request = create_printer_subscription_form.render(kw)
         response = self._call_ipptool(uri, request)
         return response['Tests'][0]
@@ -431,6 +490,24 @@ class IPPToolWrapper(object):
         response = self._call_ipptool(uri, request)
         return response['Tests'][0]
 
+    def get_notifications(self,
+                          uri,
+                          printer_uri=None,
+                          notify_subscription_ids=None,
+                          requesting_user_name=colander.null,
+                          notify_sequence_numbers=colander.null,
+                          notify_wait=colander.null):
+        kw = {'header':
+              {'operation_attributes':
+                  {'printer_uri': printer_uri,
+                   'requesting_user_name': requesting_user_name,
+                   'notify_subscription_ids': notify_subscription_ids,
+                   'notify_sequence_numbers': notify_sequence_numbers,
+                   'notify_wait': notify_wait}}}
+        request = get_notifications_form.render(kw)
+        response = self._call_ipptool(uri, request)
+        return response['Tests'][0]
+
     def cancel_subscription(self, uri,
                             printer_uri=None,
                             requesting_user_name=colander.null,
