@@ -1,7 +1,15 @@
 import colander
-from .widgets import (IPPAttributeWidget, IPPBodyWidget, IPPDisplayWidget,
-                      IPPGroupWidget, IPPNameWidget, IPPTupleWidget,
-                      IPPConstantTupleWidget)
+from .widgets import (IPPAttributeWidget, IPPBodyWidget, IPPGroupWidget,
+                      IPPNameWidget, IPPTupleWidget, IPPConstantTupleWidget)
+
+
+class IntegerOrTuple(colander.List):
+    def serialize(self, node, appstruct):
+        if isinstance(appstruct, (tuple, list)):
+            return super(IntegerOrTuple, self).serialize(
+                node,
+                ','.join([str(i) for i in appstruct]))
+        return super(IntegerOrTuple, self).serialize(node, appstruct)
 
 
 class StringOrTuple(colander.String):
@@ -20,6 +28,10 @@ class Enum(colander.String):
     pass
 
 
+class Integer(IntegerOrTuple):
+    pass
+
+
 class Keyword(StringOrTuple):
     pass
 
@@ -96,7 +108,7 @@ class SubscriptionOperationAttributes(OperationAttributesWithPrinterUri):
                                                widget=IPPAttributeWidget())
 
 
-class CreateJobSubscriptionsOperationAttributes(
+class CreateJobSubscriptionOperationAttributes(
         SubscriptionOperationAttributes):
     notify_job_id = colander.SchemaNode(colander.Integer(),
                                         widget=IPPAttributeWidget())
@@ -149,6 +161,21 @@ class GetSubscriptionsAttributes(OperationAttributesWithPrinterUri):
         widget=IPPAttributeWidget())
 
 
+class GetNotificationsAttributes(OperationAttributesWithPrinterUri):
+    requesting_user_name = colander.SchemaNode(
+        Name(),
+        widget=IPPAttributeWidget())
+    notify_subscription_ids = colander.SchemaNode(
+        Integer(),
+        widget=IPPAttributeWidget())
+    notify_sequence_numbers = colander.SchemaNode(
+        Integer(),
+        widget=IPPAttributeWidget())
+    notify_wait = colander.SchemaNode(
+        colander.Boolean(false_val=0, true_val=1),
+        widget=IPPAttributeWidget())
+
+
 class CupsGetPPDsSchemaOperationAttributes(GetDevicesOperationAttributes):
     ppd_make = colander.SchemaNode(Text(), widget=IPPAttributeWidget())
     ppd_make_and_model = colander.SchemaNode(
@@ -353,13 +380,11 @@ class CupsRejectJobsSchema(BaseIPPSchema):
         widget=IPPAttributeWidget())
 
 
-class CreateJobSubscriptionsSchema(BaseIPPSchema):
-    name = 'Create Job Subscriptions'
-    operation = 'Create-Job-Subscriptions'
-    object_attributes_tag = 'subscription-attributes-tag'
+class CreateSubscriptionSchema(BaseIPPSchema):
     header = HeaderIPPSchema(widget=IPPConstantTupleWidget())
-    header['operation_attributes'] = CreateJobSubscriptionsOperationAttributes(
+    header['operation_attributes'] = SubscriptionOperationAttributes(
         widget=IPPTupleWidget())
+    object_attributes_tag = 'subscription-attributes-tag'
     notify_recipient_uri = colander.SchemaNode(Uri(),
                                                widget=IPPAttributeWidget())
     notify_pull_method = colander.SchemaNode(Keyword(),
@@ -369,33 +394,26 @@ class CreateJobSubscriptionsSchema(BaseIPPSchema):
     notify_attributes = colander.SchemaNode(Keyword(),
                                             widget=IPPAttributeWidget())
     notify_charset = colander.SchemaNode(Charset(),
-                                         default='utf-8',
                                          widget=IPPAttributeWidget())
     notify_natural_language = colander.SchemaNode(Language(),
-                                                  default='en',
                                                   widget=IPPAttributeWidget())
     notify_time_interval = colander.SchemaNode(colander.Integer(),
                                                widget=IPPAttributeWidget())
 
 
-class CreatePrinterSubscriptionSchema(BaseIPPSchema):
+class CreateJobSubscriptionSchema(CreateSubscriptionSchema):
+    name = 'Create Job Subscription'
+    operation = 'Create-Job-Subscription'
+    header = HeaderIPPSchema(widget=IPPConstantTupleWidget())
+    header['operation_attributes'] = CreateJobSubscriptionOperationAttributes(
+        widget=IPPTupleWidget())
+
+
+class CreatePrinterSubscriptionSchema(CreateSubscriptionSchema):
     name = 'Create Printer Subscription'
     operation = 'Create-Printer-Subscription'
-    object_attributes_tag = 'subscription-attributes-tag'
-    header = HeaderIPPSchema(widget=IPPConstantTupleWidget())
-    header['operation_attributes'] = SubscriptionOperationAttributes(
-        widget=IPPTupleWidget())
-    notify_recipient_uri = colander.SchemaNode(Uri(),
-                                               widget=IPPAttributeWidget())
-    notify_events = colander.SchemaNode(Keyword(),
-                                        widget=IPPAttributeWidget())
     notify_lease_duration = colander.SchemaNode(colander.Integer(),
                                                 widget=IPPAttributeWidget())
-    notify_lease_expiration_time = colander.SchemaNode(
-        colander.Integer(),
-        widget=IPPAttributeWidget())
-    notify_subscription_id = colander.SchemaNode(colander.String(),
-                                                 widget=IPPDisplayWidget())
 
 
 class GetJobAttributesSchema(BaseIPPSchema):
@@ -434,6 +452,15 @@ class GetSubscriptionsSchema(BaseIPPSchema):
     object_attributes_tag = colander.null
 
 
+class GetNotificationsSchema(BaseIPPSchema):
+    name = 'Get Notifications'
+    operation = 'Get-Notifications'
+    header = HeaderIPPSchema(widget=IPPConstantTupleWidget())
+    header['operation_attributes'] = GetNotificationsAttributes(
+        widget=IPPTupleWidget())
+    object_attributes_tag = colander.null
+
+
 class PausePrinterSchema(BaseIPPSchema):
     name = 'Pause Printer'
     operation = 'Pause-Printer'
@@ -478,7 +505,7 @@ cancel_job_schema = CancelJobSchema(widget=IPPBodyWidget())
 
 release_job_schema = ReleaseJobSchema(widget=IPPBodyWidget())
 
-create_job_subscriptions_schema = CreateJobSubscriptionsSchema(
+create_job_subscription_schema = CreateJobSubscriptionSchema(
     widget=IPPBodyWidget())
 
 create_printer_subscription_schema = CreatePrinterSubscriptionSchema(
@@ -517,6 +544,8 @@ get_printer_attributes_schema = GetPrinterAttributesSchema(
 
 get_subscriptions_schema = GetSubscriptionsSchema(widget=IPPBodyWidget())
 
+get_notifications_schema = GetNotificationsSchema(widget=IPPBodyWidget())
+
 pause_printer_schema = PausePrinterSchema(widget=IPPBodyWidget())
 
 resume_printer_schema = ResumePrinterSchema(widget=IPPBodyWidget())
